from typing import Callable
from asyncio import gather
from asyncio import iscoroutinefunction as is_async, iscoroutine as is_awaiting
from enum import Enum
from .operators import OperatorSignatures


class MonadSignatures(Enum):
    UNWRAP = "unwrap"
    RESOLVE = "resolve"


class Monad:
    # Data
    __d = []

    # NOTE: function-chain
    __c = []

    # NOTE: async or not
    __as = False

    def __init__(self, data=None, cb=None):
        if data is not None:
            self.__d = data
            if not isinstance(data, list):
                self.__d = [data]

        self.__c = [] if not cb else cb

    def pending(self):
        return self.__c

    def __or__(self, args):
        """Operator-overloading for binding"""
        if isinstance(args, MonadSignatures):
            attr = f"async_{args.value}" if self.__as else args.value
            return getattr(self, attr)()

        next_md = Monad(data=self.__d, cb=[*self.__c, args])
        next_md.__as = self.__as or args[2]
        return next_md

    def __str__(self):
        return f"Monad[async={self.__as}]<{self.__d}>"

    def resolve(self):
        """resolve all of function chain"""
        return Monad(data=self.__resolve_chain())

    async def async_resolve(self):
        """resolve all of function chain"""
        data = await self.__async_resolve_chain()
        return Monad(data=data)

    def unwrap(self, callback: Callable = None):
        """resolve and return value"""
        if not self.__c:
            return self.__d if not callback else callback(self.__d)
        return self.__resolve_chain()

    async def __async_resolve_chain(self):
        cursor = 0
        result = [*self.__d]
        cbs = self.__c

        while cbs[cursor:]:
            sig, func, awaiting = cbs[cursor:][0]

            if sig == OperatorSignatures.JOIN:
                result = func(result)

            if sig == OperatorSignatures.VALIDATE:
                result = func(result)

            if sig == OperatorSignatures.MAP:
                result = (
                    list(map(func, result))
                    if not awaiting
                    else [await func(x) for x in result]
                )

            if sig == OperatorSignatures.FILTER:
                result = (
                    list(filter(func, result))
                    if not awaiting
                    else [x for x in result if bool(await func(x))]
                )

            if sig == OperatorSignatures.CONCAT:
                result += func(result)

            if sig == OperatorSignatures.FOLD:
                result = [func(result)] if not awaiting else [await func(result)]

            if sig == OperatorSignatures.BIND:
                result = func(result) if not awaiting else await func(result)

            if sig == OperatorSignatures.FLATTEN:
                result = func(result)

            if sig == OperatorSignatures.DISTINCT:
                result = func(result)

            if sig == OperatorSignatures.SORT:
                result = func(result)

            cursor += 1

        return result

    def __resolve_chain(self):
        cursor = 0
        result = [*self.__d]
        cbs = self.__c

        if not cbs:
            return self.__d

        while cbs[cursor:]:
            sig, func, _ = cbs[cursor:][0]

            if sig == OperatorSignatures.JOIN:
                result = func(result)

            if sig == OperatorSignatures.VALIDATE:
                result = func(result)

            if sig == OperatorSignatures.MAP:
                result = list(map(func, result))

            if sig == OperatorSignatures.FILTER:
                result = list(filter(func, result))

            if sig == OperatorSignatures.CONCAT:
                result += func(result)

            if sig == OperatorSignatures.FOLD:
                result = [func(result)]

            if sig == OperatorSignatures.BIND:
                result = func(result)

            if sig == OperatorSignatures.FLATTEN:
                result = func(result)

            if sig == OperatorSignatures.DISTINCT:
                result = func(result)

            if sig == OperatorSignatures.SORT:
                result = func(result)

            cursor += 1

        return result



from asyncio import iscoroutinefunction as is_async
from typing import Callable, List, Any
from functools import reduce
from enum import Enum


class OperatorSignatures(Enum):
    MAP = "map"
    FILTER = "filter"
    CONCAT = "concat"
    FOLD = "fold"
    FLATTEN = "flatten"
    DISTINCT = "distinct"
    SORT = "sort"
    BIND = "bind"
    JOIN = "join"
    VALIDATE = "validate"


class Op:
    """Operator to perform on monad-binding"""

    @staticmethod
    def JOIN(another_monad):
        def __(d: List[Any]):
            data = another_monad.unwrap()
            return [*d, *data]

        return OperatorSignatures.JOIN, __, False

    @staticmethod
    def BIND(fn, *args, **kwargs):
        def __(d: List[Any]):
            return fn(d, *args, **kwargs)

        return OperatorSignatures.BIND, __, is_async(fn)

    @staticmethod
    def MAP(fn, *args, **kwargs):
        def __(x: Any):
            return fn(x, *args, **kwargs)

        return OperatorSignatures.MAP, __, is_async(fn)

    @staticmethod
    def FILTER(fn, *args, **kwargs):
        def __(x: Any):
            return fn(x, *args, **kwargs)

        return OperatorSignatures.FILTER, __, is_async(fn)

    @staticmethod
    def FOLD(fn: Callable, default: Any):
        def __(d: List):
            return reduce(fn, d, default)

        return OperatorSignatures.FOLD, __, False

    @staticmethod
    def CONCAT(*args):
        def __(d: List):
            return [*d, *args]

        return OperatorSignatures.CONCAT, __, False

    @staticmethod
    def FLATTEN():
        def __(d: List):
            result = [item for sublist in d for item in sublist]
            return result

        return OperatorSignatures.FLATTEN, __, False

    @staticmethod
    def DISTINCT(*, key=lambda x: x):
        def __(d: List):
            found = set()
            result = []

            for item in d:
                identity = key(item)
                if identity in found:
                    continue
                result.append(item)
                found.add(identity)

            return result

        return OperatorSignatures.DISTINCT, __, False

    @staticmethod
    def SORT(*, key=lambda x: x, **kwargs):
        def __(d: List):
            return sorted(d, key=key, **kwargs)

        return OperatorSignatures.SORT, __, False

    @staticmethod
    def VALIDATE(test: Callable = None, model=None, failfast=True):
        def __(d: List):
            if not test and not model:
                return d
            result = []
            for item in d:
                test_validate, model_validate = True, True
                if test:
                    test_validate = test(item)
                if model:
                    model_validate = isinstance(item, model)
                if test_validate and model_validate:
                    result.append(item)
                    continue
                if failfast:
                    raise TypeError
            return result

        return OperatorSignatures.VALIDATE, __, False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mithrandir']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['lint = scripts:lint', 'test = scripts:test']}

setup_kwargs = {
    'name': 'mithrandir',
    'version': '1.1.1',
    'description': '',
    'long_description': '---\ndescription: Reasoning the logic flow with Magic\n---\n\n# Mithrandir\n\n## Installation\n\nAs much simple as it could be, use `pip` , `pipenv` or `poetry` :\n\n```\n$ pip install mithrandir\n```\n\n{% hint style="info" %}\nRemember to check out the latest version available\n{% endhint %}\n\nOnce you have **Mithrandir** in your presence, let him be your guide...\n\n{% code title="$ python" %}\n```python\n>> from mithrandir import Monad, Op, MonadSignature as Sig\n>> magic = Monad(1) | Op.MAP(lambda x: x + 1)\n>> print("magic = ", magic)\n... "magic = Monad<async_mode=False>[2]"\n>> # Give me your blessing...\n>> print("blessed-value = ", magic.unwrap())\n... "blessed-value = 2"\n```\n{% endcode %}\n\nFor more advanced tutorials and examples, check out the next parts of the document\n\n',
    'author': 'vutr',
    'author_email': 'me@vutr.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
